! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: badpixutils_mod
!
! Authors:  J. Ballet - SAp  CEA Saclay
!          
! Description: Generic routines to deal with bad pixels
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module badpixutils_mod

  use emutils_mod
  use cal_mos
  use datafilter  
  use dal
  use errorHandling

  implicit none 

  private

  public :: readBadpix, writeBadpix, mergeBad, localMedian, readBadOffsets
  public :: cumulativePoisson, compCumulPoisson, quantilePoisson, corrCoeff
  public :: cumulativeBinomial

  interface localMedian
     module procedure intmedloc
     module procedure realmedloc
  end interface

contains

! Write bad pixels file

  subroutine writeBadpix(bad_tab, xbad, ybad, yext, tbad, fbad, nbad)

! bad_tab : Handle to the output bad pixels table
! xbad    : array of RAWX coordinates
! ybad    : array of RAWY coordinates
! yext    : array of RAWY extensions
! tbad    : array of bad pixel types
! fbad    : array of bad pixel status (uplinked, CCF or new)
! nbad    : number of bad pixels

    type(TableT),                      intent(in) :: bad_tab
    integer(kind=int16), dimension(:), intent(in) :: xbad, ybad, yext,  &
                                                     tbad, fbad
    integer,                           intent(in) :: nbad

    type(ColumnT)  :: badx_col, bady_col, badt_col, bade_col, badf_col
    integer(kind=int16), dimension(:), pointer :: badx, bady, badtype,  &
                                                  badext, badflag
    integer :: nrows

    write(errstr,*) "Writing", nbad, " rows to ", trim(qualifiedName(bad_tab))
    call message(AppMsg,Sparse)

    nrows = numberOfRows(bad_tab)
    if (nbad > nrows) then
       call insertRows(bad_tab, nrows, nbad-nrows)
    else if (nbad < nrows) then
       call deleteRows(bad_tab, nbad, nrows-nbad)
    endif

    if (nbad > 0) then
       if ( .not. hasAttribute(bad_tab,"INSTRUME") ) then
            badx_col = column(bad_tab, "RAWX", Modify)
            bady_col = column(bad_tab, "RAWY", Modify)
       else if ( (trim(stringAttribute(bad_tab,"INSTRUME")) .eq. 'RGS1') .or. &
            (trim( stringAttribute(bad_tab,"INSTRUME") ) .eq. 'RGS2') ) then
            badx_col = column(bad_tab, "CHIPX", Modify)
            bady_col = column(bad_tab, "CHIPY", Modify)
       else
            badx_col = column(bad_tab, "RAWX", Modify)
            bady_col = column(bad_tab, "RAWY", Modify)
       endif

       badt_col = column(bad_tab, "TYPE", Modify)
       bade_col = column(bad_tab, "YEXTENT", Modify)
       badf_col = column(bad_tab, "BADFLAG", Modify)

       badx    => int16Data(badx_col)
       bady    => int16Data(bady_col)
       badtype => int16Data(badt_col)
       badext  => int16Data(bade_col)
       badflag => int16Data(badf_col)

       badtype = tbad(1:nbad)
       badflag = fbad(1:nbad)
       badext  = yext(1:nbad)
       badx    = xbad(1:nbad)
       bady    = ybad(1:nbad)
    endif

  end subroutine writeBadpix

!! Read bad pixels file

  subroutine readBadpix(bad_tab, incremental,  &
                        xbad, ybad, yext, tbad, fbad, nbad)

! bad_tab : Handle to the bad pixels table
! incremental : Add new columns to bad pixels file if false
! xbad    : array of RAWX coordinates
! ybad    : array of RAWY coordinates
! yext    : array of RAWY extensions
! tbad    : array of bad pixel types
! fbad    : array of bad pixel status (uplinked or not)
! nbad    : number of bad pixels

    type(TableT),                      intent(in)  :: bad_tab
    logical,                           intent(in)  :: incremental
    integer(kind=int16), dimension(:), intent(out) :: xbad, ybad, yext,  &
                                                      tbad, fbad
    integer,                           intent(out) :: nbad

    type(ColumnT)  :: badx_col, bady_col, badt_col, bade_col, badf_col
    integer(kind=int16), dimension(:), pointer :: badx, bady, badtype,  &
                                                  badext, badflag
    logical :: withflag

    withflag = .true.
    if (incremental) then

       nbad = numberOfRows(bad_tab)
       if (nbad <= size(xbad)) then
          write(errstr,*) "Found", nbad, " bad pixels in ",  &
                          trim(qualifiedName(bad_tab))
          call message(AppMsg,Sparse)
       else
          write(errstr,*) "The number", nbad, " of bad pixels in ",  &
                          trim(qualifiedName(bad_tab)), " is too large ",  &
                          "(maximum", size(xbad), "). Contact developer."
          call error("readBadpix01")
       endif

       if ( .not. hasAttribute(bad_tab,"INSTRUME") ) then
            badx_col = column(bad_tab, "RAWX", Modify)
            bady_col = column(bad_tab, "RAWY", Modify)
       else if ( (trim(stringAttribute(bad_tab,"INSTRUME")) .eq. 'RGS1') .or. &
        (trim( stringAttribute(bad_tab,"INSTRUME") ) .eq. 'RGS2') ) then
            badx_col = column(bad_tab, "CHIPX", Read)
            bady_col = column(bad_tab, "CHIPY", Read)
       else
            badx_col = column(bad_tab, "RAWX", Read)
            bady_col = column(bad_tab, "RAWY", Read)
       endif

       badt_col = column(bad_tab, "TYPE", Read)
       bade_col = column(bad_tab, "YEXTENT", Read)
       badx    => int16Data(badx_col)
       bady    => int16Data(bady_col)
       badtype => int16Data(badt_col)
       badext  => int16Data(bade_col)
       tbad(1:nbad) = badtype
       yext(1:nbad) = badext
       xbad(1:nbad) = badx
       ybad(1:nbad) = bady

! The BADFLAG column is not created by badpixfind. Set it to 3 (file)

       withflag = hasColumn(bad_tab, "BADFLAG")
       if (withflag) then
          badf_col = column(bad_tab, "BADFLAG", Read)
          badflag => int16Data(badf_col)
          fbad(1:nbad) = badflag
       else
          fbad(1:nbad) = 3
       endif

    else
       if ( .not. hasAttribute(bad_tab,"INSTRUME") ) then
            badx_col = addColumn(bad_tab, "RAWX", Integer16)
            bady_col = addColumn(bad_tab, "RAWY", Integer16)
        else if  ( trim(stringAttribute(bad_tab,"INSTRUME")) .eq. 'RGS1' .or. &
        trim( stringAttribute(bad_tab,"INSTRUME") ) .eq. 'RGS2')   then
            badx_col = addColumn(bad_tab, "CHIPX", Integer16)
            bady_col = addColumn(bad_tab, "CHIPY", Integer16)
       else
            badx_col = addColumn(bad_tab, "RAWX", Integer16)
            bady_col = addColumn(bad_tab, "RAWY", Integer16)

       endif
       badt_col = addColumn(bad_tab, "TYPE", Integer16)
       bade_col = addColumn(bad_tab, "YEXTENT", Integer16)
       badf_col = addColumn(bad_tab, "BADFLAG", Integer16)
       nbad = 0
    endif

    call release(badx_col)
    call release(bady_col)
    call release(badt_col)
    call release(bade_col)
    if (withflag)  call release(badf_col)

  end subroutine readBadpix

! Compute Y extent of bad pixels

  subroutine mergeBad(xbad, ybad, yext, tbad, fbad, nbad)

! xbad     : array of RAWX coordinates
! ybad     : array of RAWY coordinates
! yext     : array of RAWY extensions
! tbad     : array of bad pixel types
! fbad     : array of bad pixel status (uplinked, CCF or new)
! nbad     : number of bad pixels

    integer(kind=int16), dimension(:), intent(inout) :: xbad, ybad, yext,  &
                                                        tbad, fbad
    integer,                           intent(inout) :: nbad

! Order bad pixel types by increasing gravity
! HOT(1) > FLICKERING(2) > PIN_HOLE(4) > DEAD(3) > UNSPECIFIED(5) > INTACT(0)

    integer, dimension(0:5), parameter :: gravity = (/0,5,4,2,3,1/)

    integer, dimension(:), allocatable :: xbad2, ybad2, yext2, tbad2, fbad2,  &
                                          perm
    integer :: nbad2, ier, k1, k2, m1, m2, m3, ymin, ymax, order1, order2
    logical :: recheck

    if (nbad <= 1)  return

! The output may be larger if overlap exists

    allocate(xbad2(2*nbad), ybad2(2*nbad), yext2(2*nbad), tbad2(2*nbad))
    allocate(perm(2*nbad), fbad2(2*nbad))
    recheck = .true.

    do while (recheck)
       recheck = .false.

! Sort bad pixels per column then per row then per decreasing type gravity
! and increasing flag

       ymin = minval(ybad(1:nbad))
       ymax = maxval(ybad(1:nbad)+yext(1:nbad)-1)
       ybad2(1:nbad) = ((xbad(1:nbad)*(ymax-ymin+1) + ybad(1:nbad)-ymin)*10  &
                     - gravity(tbad(1:nbad)))*10 + fbad(1:nbad)
       call ipsort(ybad2, nbad, perm, 1, ier)
       if (ier > 0) then
          write(errstr,*) "error in call to ipsort : ier =", ier
          call error("mergeBad01")
       endif
       perm(nbad+1) = 1

! Loop on bad pixels

       k1 = 1
       nbad2 = 0
       do while (nbad2 < 2*nbad)
          m1 = perm(k1)
          ymax = ybad(m1) + yext(m1)

! Look for all bad pixels vertically adjacent to that one
! Define order of precedence using type and flag (uplinked > CCF > new)

          order1 = gravity(tbad(m1))*10 - fbad(m1)
          do k2 = k1+1, nbad
             m2 = perm(k2)
             if (xbad(m2) > xbad(m1) .or. ybad(m2) > ymax)  exit
             order2 = gravity(tbad(m2))*10 - fbad(m2)
             if (order2 == order1) then

! Simply merge bad pixels and continue

                ymax = max(ymax,int(ybad(m2)+yext(m2)))
             else if (order2 < order1) then

! Erase the coincident part of m2. Set it aside for new loop
! if yet another bad pixel interferes

                if (ybad(m2)+yext(m2) > ymax) then
                   yext(m2) = yext(m2) + ybad(m2) - ymax
                   ybad(m2) = ymax
                   m3 = perm(k2+1)
                   if (xbad(m3) == xbad(m1) .and. ybad(m3) < ymax .and.  &
                       k2 < nbad ) then
                      recheck = .true.
                      nbad2 = nbad2 + 1
                      xbad2(nbad2) = xbad(m2)
                      ybad2(nbad2) = ybad(m2)
                      yext2(nbad2) = yext(m2)
                      tbad2(nbad2) = tbad(m2)
                      fbad2(nbad2) = fbad(m2)
                   else
                      exit
                   endif
                endif

! Final case where the second bad pixel supersedes the first
! Thanks to the initial ordering it is not possible to have equal ybad here

             else if (ybad(m2)+yext(m2) >= ymax) then
                ymax = ybad(m2)
                exit
             else

! Replace the coincident part of m1 and split it
! Set it aside for new loop if yet another bad pixel interferes

                nbad2 = nbad2 + 1
                xbad2(nbad2) = xbad(m1)
                ybad2(nbad2) = ybad(m1)
                yext2(nbad2) = ybad(m2) - ybad(m1)
                tbad2(nbad2) = tbad(m1)
                fbad2(nbad2) = fbad(m1)
                ybad(m1) = ybad(m2) + yext(m2)
                m3 = perm(k2+1)
                if (xbad(m3) == xbad(m1) .and. ybad(m3) <= ybad(m2)+yext(m2)  &
                    .and. k2 < nbad ) then
                   recheck = .true.
                   exit
                else
                   nbad2 = nbad2 + 1
                   xbad2(nbad2) = xbad(m2)
                   ybad2(nbad2) = ybad(m2)
                   yext2(nbad2) = yext(m2)
                   tbad2(nbad2) = tbad(m2)
                   fbad2(nbad2) = fbad(m2)
                endif
             endif
          enddo
          nbad2 = nbad2 + 1
          xbad2(nbad2) = xbad(m1)
          ybad2(nbad2) = ybad(m1)
          yext2(nbad2) = ymax - ybad(m1)
          tbad2(nbad2) = tbad(m1)
          fbad2(nbad2) = fbad(m1)
          if (k2 > nbad)  exit
          k1 = k2
       enddo

! Replace routine arguments

       if (nbad2 > size(xbad)) then
          write(errstr,*) "Output of mergeBad is larger", nbad2,  &
                          " than array size", size(xbad),  &
                          ". Developer of calling program should beware."
          call error("mergeBad02")
       endif
       nbad = nbad2
       xbad(1:nbad) = xbad2(1:nbad)
       ybad(1:nbad) = ybad2(1:nbad)
       yext(1:nbad) = yext2(1:nbad)
       tbad(1:nbad) = tbad2(1:nbad)
       fbad(1:nbad) = fbad2(1:nbad)
    enddo
    deallocate(xbad2, ybad2, yext2, tbad2, fbad2, perm)

  end subroutine mergeBad

! Compute cumulative binomial distribution
! cumulativeBinomial(Non,Noff,p) = Sum(Non to Non+Noff) PB(Non,Noff,p)
! PB(Non,Noff,p) is the probability to get Non source counts and Noff
! background counts, if p is the a priori probability that a count is
! attributed to the source (on assumption of no source)

  real(double) function cumulativeBinomial(Non, Noff, p)

! Input:
! Non  : number of observed source counts
! Noff : number of observed background counts
! p    : a priori probability that a count is attributed to the source

    real(double), intent(in) :: p
    integer,      intent(in) :: Non, Noff

    real(double), external :: dbetai
    real(double)   :: cur, cumul, ratio
    integer        :: k, Ntot

    if (p < 10*tiny(p)) then
       write(errstr,*) "Probability", p, " is too small."
       call error("cumulativeBinomial02")
    endif

    Ntot = Non + Noff
    if (Ntot > 12) then

! Use incomplete beta function

       cumulativeBinomial = dbetai(p,real(Non,double),real(Noff+1,double))
    else

! Compute cumulative distribution iteratively, starting at 0 background count

       cur = p**Ntot
       cumul = cur
       ratio = 1.d0/p - 1.d0
       do k = 1, Noff
          cur = cur * ratio * (Ntot-k+1) / k
          cumul = cumul + cur
       enddo
       cumulativeBinomial = cumul
    endif

  end function cumulativeBinomial

! Compute cumulative Poisson distribution over some range
! cumulativePoisson(k) = Sum(0 to k) P(k)

  subroutine cumulativePoisson(mu, kmin, kmax, cvf)

! Input:
! mu      : average value
! kmin    : minimum number of counts
! kmax    : maximum number of counts
! Output:
! cvf(1:kmax-kmin+1) : cumulative Poisson distribution from kmin to kmax

    real(double), intent(in) :: mu
    integer,      intent(in) :: kmin, kmax
    real(double), dimension(:), intent(out) :: cvf

    real(double), external :: derf, derfc
    real(double)   :: cur, cumul
    integer        :: k, kmin2

    if (size(cvf) < kmax-kmin+1) then
       write(errstr,*) "Size of output array", size(cvf),  &
                       " should be at least kmax-kmin+1 =", kmax-kmin+1
       call error("cumulativePoisson01")
    endif

    if (mu < 10*tiny(mu)) then
       write(errstr,*) "Average value", mu, " is too small."
       call error("cumulativePoisson02")
    endif

    if (mu >= 300) then

! Use Gauss limit for Poisson distribution
! Limit for small probabilities

       kmin2 = max(int(mu-3*sqrt(mu))+1,kmin) - 1
       do k = kmin, kmin2
          cvf(k-kmin+1) = derfc((mu-k-0.5d0)/sqrt(2*mu)) / 2.d0
       enddo
       do k = kmin2+1, kmax
          cvf(k-kmin+1) = (derf((k-mu+0.5d0)/sqrt(2*mu)) + 1.d0) / 2.d0
       enddo
    else

! Compute cumulative distribution iteratively

       cur = 1.d0
       cumul = cur
       if (kmin == 0)  cvf(1) = cumul
       do k = 1, kmax
          cur = cur * mu / k
          cumul = cumul + cur
          if (k >= kmin)  cvf(k-kmin+1) = cumul
       enddo
       cvf = cvf * exp(-mu)
    endif

  end subroutine cumulativePoisson

! Compute complementary cumulative Poisson distribution over some range
! compCumulPoisson(k) = Sum(k to infinity) P(k)

  subroutine compCumulPoisson(mu, kmin, kmax, cvf)

! Input:
! mu      : average value
! kmin    : minimum number of counts
! kmax    : maximum number of counts
! Output:
! cvf(1:kmax-kmin+1) : complementary cumulative Poisson distribution
!                      from kmin to kmax

    real(double), intent(in) :: mu
    integer,      intent(in) :: kmin, kmax
    real(double), dimension(:), intent(out) :: cvf

    real(double), external :: derf, derfc
    real(double)   :: cur, cumul
    integer        :: k, kmax2

    if (size(cvf) < kmax-kmin+1) then
       write(errstr,*) "Size of output array", size(cvf),  &
                       " should be at least kmax-kmin+1 =", kmax-kmin+1
       call error("compCumulPoisson01")
    endif

    if (mu < 10*tiny(mu)) then
       write(errstr,*) "Average value", mu, " is too small."
       call error("compCumulPoisson02")
    endif

! Limit for small probabilities

    kmax2 = min(int(mu+3*sqrt(mu)),kmax) + 1

    if (mu >= 300) then

! Use Gauss limit for Poisson distribution

       do k = kmin, kmax2-1
          cvf(k-kmin+1) = (derf((mu-k+0.5d0)/sqrt(2*mu)) + 1.d0) / 2.d0
       enddo
       do k = kmax2, kmax
          cvf(k-kmin+1) = derfc((k-mu-0.5d0)/sqrt(2*mu)) / 2.d0
       enddo
    else

! Compute cumulative distribution iteratively
! Use incomplete gamma function for small probabilities
! gammp(k,mu) = Sum(k to infinity) P(k,mu)

       cur = 1.d0
       cumul = cur
       if (kmin == 0)  cvf(1) = 0.d0
       if (kmin <= 1)  cvf(1-kmin+1) = cumul
       do k = 1, kmax2-1
          cur = cur * mu / k
          cumul = cumul + cur
          if (k >= kmin-1)  cvf(k-kmin+2) = cumul
       enddo
       cvf(1:kmax2-kmin+1) = 1.d0 - cvf(1:kmax2-kmin+1) * exp(-mu)
       do k = kmax2, kmax
          cvf(k-kmin+1) = exp(lgammp(real(k),real(mu)))
       enddo
    endif

  end subroutine compCumulPoisson

! Compute quantiles for the Poisson distribution
! Probability to get quantile or less is always >= 1 - epsilon
! Probability to get 1+quantile or more is always < epsilon
! gaussCvf must be a separate F77 routine to be accessible by fzero
! Get lower quantile if epsilon < 0
! such that probability to get quantile or more is always >= 1 - epsilon
! Probability to get quantile-1 or less is always < epsilon

  integer function quantilePoisson(mu, epsilon)

! mu      : average value
! epsilon : probability level

    real, intent(in) :: mu, epsilon

    real, external :: gaussCvf
    real(double)   :: target, mud, cur, cumul
    real           :: b, c, r
    integer        :: ier
    logical        :: negative

    if (abs(epsilon) >= 0.5) then
       write(errstr,*) "epsilon larger than 0.5:", epsilon,  &
                       " Contact developer."
       call error("quantilePoisson02")
    else if (abs(epsilon) < 10*tiny(epsilon)) then
       write(errstr,*) "epsilon =", epsilon, " is too small."
       call error("quantilePoisson03")
    endif

    negative = epsilon < 0.

    if (mu >= 300) then

! Initialize gaussCvf

       b = gaussCvf(-abs(epsilon))

! Use Gauss limit for Poisson distribution (rather bad approximation)
! Find lower and upper limit to threshold

       if (gaussCvf(1.) > 0) then
          b = max(sqrt(-log(3.*abs(epsilon))/2.),1.)
          c = sqrt(-log(2.*abs(epsilon))*2.)
       else
          b = 0.
          c = 1.
       endif
       r = c
       call fzero(gaussCvf, b, c, r, 1.e-5, 1.e-5, ier)
       if (ier >= 3) then
          write(errstr,*) "error in call to fzero: iflag =", ier,  &
                          " Contact developer."
          call error("quantilePoisson01")
       endif
       if (negative) then
          quantilePoisson = int(mu - sqrt(mu)*b + 0.5)
       else
          quantilePoisson = int(mu + sqrt(mu)*b + 0.5)
       endif

! Use incomplete gamma function for small epsilon
! gammp(k,mu) = Sum(k to infinity) P(k,mu)

    else if (.not. negative .and. epsilon < 10.**(0.5-precision(1.0))) then
       quantilePoisson = int(mu) + 1
       do while (lgammp(real(quantilePoisson+1),mu) >= log(epsilon))
          quantilePoisson = quantilePoisson + 1
       enddo
    else

! Compute cumulative distribution iteratively

       mud = real(mu,kind=double)
       if (negative) then
          target = real(-epsilon,kind=double) * exp(mud)
       else
          target = (1.d0 - real(epsilon,kind=double)) * exp(mud)
       endif
       quantilePoisson = 0
       cur = 1.d0
       cumul = cur
       do while (cumul <= target)
          quantilePoisson = quantilePoisson + 1
          cur = cur * mud / quantilePoisson
          cumul = cumul + cur
       enddo
    endif

  end function quantilePoisson

! Return log of normalised incomplete gamma function
! (Int(0,x) exp(-t) t^(a-1) dt) / Gamma(a)

  real function lgammp(a,x)

    real, intent(in) :: a, x

! Use incomplete gamma function from slatec

    real(double), external :: d9lgit, dlngam
    real(double) :: da,dx

    da = real(a,kind=double)
    dx = real(x,kind=double)
    lgammp = d9lgit(da,dx,dlngam(da+1.d0)) + da * log(dx)

  end function lgammp

! Return Pearson's correlation coefficient
! Return 1 if either all X or all Y values are equal

  real function corrCoeff(x,y,theta,sigma1,sigma2)

! x, y   : list of X and Y values
! Returns if present:
! theta  : rotation angle (radians, between -pi/4 and pi/4) to main axes
! sigma1 : dispersion along theta (not necessarily major axis)
! sigma2 : dispersion perpendicular to theta

    real, intent(in) , dimension(:) :: x, y
    real, intent(out), optional     :: theta, sigma1, sigma2

    real(double) :: avx, avy, vx, vy, vxy, theta2
    integer      :: npts

    npts = size(x)

    if (size(y) /= npts) then
       write(errstr,*) "Size of X and Y lists are not the same", npts, size(y)
       call error("corrCoeff01")
    endif

! Compute averages

    avx = sum(real(x,kind=double)) / npts
    avy = sum(real(y,kind=double)) / npts

! Compute variances

    vx  = sum((real(x,kind=double)-avx)**2)
    vy  = sum((real(y,kind=double)-avy)**2)
    vxy = sum((real(x,kind=double)-avx)*(real(y,kind=double)-avy))

    if (vx < 10*tiny(vx) .or. vy < 10*tiny(vy)) then
       corrcoeff = 1.
    else
       corrCoeff = vxy / sqrt(vx) / sqrt(vy)
    endif

    if (abs(vx-vy) < 10*tiny(vx)) then
       theta2 = 0.5d0 * acos(0.d0)
    else
       theta2 = 0.5d0 * atan(2.d0*vxy/(vx-vy))
    endif
    if (present(theta))  &
         theta = theta2   ! theta2 is double precision
    if (present(sigma1)) &
         sigma1 = sqrt(cos(theta2)**2*vx + sin(2.d0*theta2)*vxy +  &
                       sin(theta2)**2*vy)
    if (present(sigma2)) &
         sigma2 = sqrt(sin(theta2)**2*vx - sin(2.d0*theta2)*vxy +  &
                       cos(theta2)**2*vy)

  end function corrCoeff

! Compute median of an array

  integer function intmedloc(toto,nval)

! toto: 1-D input array of integer values
! nval: number of values in toto to consider (optional)

    integer, dimension(:), intent(in)  :: toto
    integer, optional,     intent(in)  :: nval

    integer, dimension(:), allocatable :: tata, tutu
    integer :: nelem

    nelem = size(toto)
    if (present(nval)) then
       if (nval > nelem) then
          write(errstr,*) "nval =",nval," > size(toto) =",nelem
          call error("medloc01")
       else if (nval < 1) then
          write(errstr,*) "nval =",nval," < 1"
          call error("medloc03")
       else
          nelem = nval
       endif
    endif
    allocate(tata(nelem),tutu(nelem))
    tutu = toto(1:nelem)
    call isort(tutu,tata,nelem,1)
    intmedloc = (tutu((nelem+1)/2) + tutu(nelem/2+1)) / 2
    deallocate(tata,tutu)

  end function intmedloc

! Compute median of an array

  real function realmedloc(toto,nval)

! toto: 1-D input array of real values
! nval: number of values in toto to consider (optional)

    real,    dimension(:), intent(in)  :: toto
    integer, optional,     intent(in)  :: nval

    real,    dimension(:), allocatable :: tata, tutu
    integer :: nelem

    nelem = size(toto)
    if (present(nval)) then
       if (nval > nelem) then
          write(errstr,*) "nval =",nval," > size(toto) =",nelem
          call error("medloc01")
       else if (nval < 1) then
          write(errstr,*) "nval =",nval," < 1"
          call error("medloc03")
       else
          nelem = nval
       endif
    endif
    allocate(tata(nelem),tutu(nelem))
    tutu = toto(1:nelem)
    call ssort(tutu,tata,nelem,1)
    realmedloc = 0.5 * (tutu((nelem+1)/2) + tutu(nelem/2+1))
    deallocate(tata,tutu)

  end function realmedloc

! Read bad offsets in SAS coordinates 1-600
! Beware: contains under/overscans

  subroutine readBadOffsets(ev_set, offX, offY, ccdnr)

! ev_set    : Handle to the data set
! offX, offY: additional offset values
! ccdnr     : CCD number as CCDNR column (if merged table)

    type(DataSetT),                        intent(in)  :: ev_set
    integer, dimension(-10:EMOS_MAX_X+20), intent(out) :: offX
    integer, dimension(-10:EMOS_MAX_Y+20), intent(out) :: offY
    integer, optional,                     intent(in)  :: ccdnr

    character(len=*), parameter :: badoff_ext = "OFFSETS"

    type(TableT)      :: off_tab
    type(ColumnT)     :: pb_col, obx_col, oby_col
    integer(kind=int16), dimension(:), pointer :: pbad, obadx, obady
    integer(kind=int8),  dimension(:), pointer :: ccdval
    logical, dimension(:), allocatable         :: goodccd
    integer           :: irow, ibad, nrows, ccdcur, ngood

    offX = 0
    offY = 0
    if (hasBlock(ev_set,badoff_ext)) then
       off_tab = table(ev_set,badoff_ext)
       nrows = numberOfRows(off_tab)

       if (nrows > 0) then
          allocate(goodccd(0:nrows-1))
          goodccd = .true.

! Is the table merged from several CCDs ?
! If so, select the required CCDNR

          if (hasColumn(off_tab,"CCDNR")) then
             if (present(ccdnr)) then
                ccdval => int8Data(column(off_tab, 'CCDNR'))
                goodccd = (ccdval == ccdnr)
             else
                write(errstr,*) trim(qualifiedName(off_tab)), " contains ",  &
                     "several CCDs. The ccdnr argument should be present."
                call error("readBadOffsets01")
             endif
          else if (present(ccdnr)) then
             ccdcur = getCcd(block(off_tab))
             if (ccdcur /= ccdnr) then
                write(errstr,*) trim(qualifiedName(off_tab)), " contains ",  &
                     "offsets for CCDNR =", ccdcur, " whereas CCDNR =", &
                     ccdnr, " was required."
                call warning("readBadOffsets10")
                goodccd = .false.
             endif
          endif

! Read offsets

          ngood = count(goodccd)
          write(errstr,*) "Found ", ngood, " matching rows in ",  &
                          trim(qualifiedName(off_tab))

          if (ngood > 0) then
             if ( .not. hasAttribute(off_tab,"INSTRUME") ) then
                    pb_col = column(off_tab,"RAWX",Read)
             else if ( (trim(stringAttribute(off_tab,"INSTRUME")) .eq. 'RGS1') .or. &
                (trim( stringAttribute(off_tab,"INSTRUME") ) .eq. 'RGS2') ) then
                    pb_col = column(off_tab,"CHIPX",Read)
             else
                    pb_col = column(off_tab,"RAWX",Read)
             endif
             obx_col = column(off_tab,"OFFSETX",Read)
             oby_col = column(off_tab,"OFFSETY",Read)
             pbad  => int16Data(pb_col)
             obadx => int16Data(obx_col)
             obady => int16Data(oby_col)

             do ibad = 0, nrows-1
                if (goodccd(ibad)) then
                   irow = pbad(ibad)
                   if (obadx(ibad) > 0) then
                      offX(irow) = obadx(ibad)
                   else
                      offY(irow) = obady(ibad)
                   endif
                endif
             enddo
          endif
          deallocate(goodccd)
       else
          write(errstr,*) "No bad offset found in ",  &
                          trim(qualifiedName(off_tab))
       endif

       call release(off_tab)
    else
       write(errstr,*) "No ", badoff_ext, " extension found in ",  &
                       trim(name(ev_set))
    endif
    call message(AppMsg,Verbose)

  end subroutine readBadOffsets

end module badpixutils_mod
