! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: emutils_mod
!
! Authors:  J. Ballet - SAp  CEA Saclay
!          
! Description: Generic routines to deal with keywords
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module emutils_mod

  use dal
  use param
  use errorHandling

  implicit none 

  private

  public :: changexy
  public :: equalKeywords, putPrimaryKeywords
  public :: sizeListParam, readListParam
  public :: keywordDone, keywordRemove, wasDone
  public :: addFilename, changeCase, getCcd, getMode

! Name of the keyword for changing raw CCD coordinates

  character(len=*), parameter :: changexy = "setcoordinates"

contains

! Write history line with the name of an input file, removing the directory

  subroutine addFilename(in_tab, filename, comment)

! in_tab    : handle to the output table
! filename  : name of the file
! comment   : comment string to introduce file name

    type(BlockT), intent(in) :: in_tab
    character(len=*), intent(in) :: filename, comment

    character(len=1024) :: comm
    integer :: enddir, lcom, lname

    enddir = index(filename,'/',.true.)
    lname = len_trim(filename)
    lcom = len_trim(comment)
    comm = comment(1:lcom) // ' ' // filename(enddir+1:lname)
    lcom = lcom + 1 + lname - enddir
    call addHistory(in_tab, comm(1:lcom))

  end subroutine addFilename

! Switch from lower to upper case or vice-versa

  function changeCase(instring,dir) result(outstring)

! instring : input string
! dir      : direction (1: to lower case; 2: to uppercase; 0: switch case)

    character(len=*), intent(in) :: instring
    integer, intent(in) :: dir
    character(len=len(instring)) :: outstring

! ASCII positions of lower case and upper case letters

    integer, parameter :: start_up  = iachar("A"), end_up  = iachar("Z"), &
                          start_low = iachar("a"), end_low = iachar("z"), &
                          offset = start_low - start_up

    integer :: length, icur, pos

    length = len(instring)
    outstring = instring
    do icur = 1, length
       pos = iachar(instring(icur:icur))
       if (pos >= start_up .and. pos <= end_up .and. dir < 2) then
          pos = pos + offset
       else if (pos >= start_low .and. pos <= end_low .and. dir /= 1) then
          pos = pos - offset
       endif
       outstring(icur:icur) = achar(pos)
    enddo

  end function changeCase

! Get the CCD value from the keywords (returns 0 if error)

  integer function getCcd(in_tab)

!  in_tab  : handle to the  input block for the DAL

    type(BlockT), intent(in) :: in_tab

! Name of the keywords for CCD and node

    character(len=*), parameter :: ccd_kw = "CCDID", node_kw = "CCDNODE",  &
                                   quad_kw = "QUADRANT"

    integer :: ccd, node, quad
    character(len=16) :: instname

    getCcd = 0
    if (hasAttribute(in_tab,ccd_kw)) then
       ccd = int32Attribute(in_tab,ccd_kw)
    else
       write(errstr,*) "No ", ccd_kw, " keyword in table ",  &
                       qualifiedName(in_tab)
       call warning("getCcd10")
       return
    endif

    if (hasAttribute(in_tab,"INSTRUME")) then
       instname = stringAttribute(in_tab,"INSTRUME")
! Fix in the case there is no INSTRUME keyword in the BADPIX extension
    else if (hasAttribute(in_tab,node_kw)) then
          instname = "EMOS1"
    else if (hasAttribute(in_tab,quad_kw)) then
          instname = "EPN"
    else
       write(errstr,*) "No INSTRUME keyword in table ", qualifiedName(in_tab)
       call warning("getCcd11")
       return
    endif

    select case(trim(instname))
    case ("RGS1", "RGS2")   ! Set node to 0 if no CCDNODE in that case
       if (ccd < 1 .or. ccd > 9) then
          write(errstr,*) "keyword ", ccd_kw, " = ", ccd,  &
             "should be between 1 and 7 in table ", qualifiedName(in_tab)
          call warning("getCcd14")
          return
       endif
       if (hasAttribute(in_tab,node_kw)) then
          node = int32Attribute(in_tab,node_kw)
          if (node < 0 .or. node > 1) then
             write(errstr,*) "keyword ", node_kw, " = ", node,  &
                "should be one of 0 or 1 in table ", qualifiedName(in_tab)
             call warning("getCcd18")
          endif
       else
          node = 0
       endif
       getCcd = 10*node + ccd
    case ("EMOS1", "EMOS2")
       if (ccd < 1 .or. ccd > 7) then
          write(errstr,*) "keyword ", ccd_kw, " = ", ccd,  &
             "should be between 1 and 7 in table ", qualifiedName(in_tab)
          call warning("getCcd14")
          return
       endif
       if (hasAttribute(in_tab,node_kw)) then
          node = int32Attribute(in_tab,node_kw)
          if (node < 0 .or. node > 1) then
             write(errstr,*) "keyword ", node_kw, " = ", node,  &
                "should be one of 0 or 1 in table ", qualifiedName(in_tab)
             call warning("getCcd18")
          else
             getCcd = 10*node + ccd
          endif
       else
          write(errstr,*) "No ", node_kw, " keyword in table",  &
                          qualifiedName(in_tab)
          call warning("getCcd12")
       endif
    case ("EPN")
       if (ccd < 0 .or. ccd > 2) then
          write(errstr,*) "keyword ", ccd_kw, " = ", ccd,  &
             "should be one of 0, 1, 2 in table ", qualifiedName(in_tab)
          call warning("getCcd13")
          return
       endif
       if (hasAttribute(in_tab,quad_kw)) then
          quad = int32Attribute(in_tab,quad_kw)
          if (quad < 0 .or. quad > 3) then
             write(errstr,*) "keyword ", quad_kw, " = ", quad,  &
                "should be between 0 and 3 in table ", qualifiedName(in_tab)
             call warning("getCcd17")
             return
          endif
          getCcd = quad*3 + ccd + 1
       else
          write(errstr,*) "No ", quad_kw, " keyword in table",  &
                          qualifiedName(in_tab)
          call warning("getCcd15")
       endif
    case default
       write(errstr,*) "keyword INSTRUME = ", trim(instname),  &
          "should be one of RGS1, RGS2, EMOS1, EMOS2, EPN in table ",  &
          qualifiedName(in_tab)
       call warning("getCcd16")
    end select

  end function getCcd

! Return logicals defining the data mode

  subroutine getMode(ev_tab, imaging, timing, redImaging, compTiming)

! ev_tab:     handle to the events extension
! imaging:    set to True if IMAGING (EPIC) or SPECTROSCOPY (RGS) mode
! timing:     set to True if TIMING (EPIC) or HTR (RGS) mode
! redImaging: set to True if REDUCED IMAGING (MOS) mode
! compTiming: set to True if COMPRESSED TIMING (MOS) or BURST (PN) mode

    type(TableT), intent(in) :: ev_tab
    logical, intent(out)     :: imaging, timing, redImaging, compTiming

    character(len=16) :: keyword

    if (.not. hasAttribute(ev_tab,"DATATYPE")) then
       write(errstr,*) "No DATATYPE keyword in table ",qualifiedName(ev_tab), &
                       ". The mode cannot be determined."
       call error("getMode02")
    endif

    keyword = stringAttribute(ev_tab,"DATATYPE")
    imaging = .false.
    timing  = .false.
    redImaging = .false.
    compTiming = .false.
    select case(trim(keyword))
    case('IMAGE.EL')
       imaging    = .true.
    case('SPECT.EL')
       imaging    = .true.
    case('RIMAGE.EL')
       redImaging = .true.
    case('TIME.EL')
       timing     = .true.
    case('HTR.EL')
       timing     = .true.
    case('CTIME.EL')
       compTiming = .true.
    case('BURST.EL')
       compTiming = .true.
    case default
       write(errstr,*) "The DATATYPE keyword ", trim(keyword), " in table ",  &
          qualifiedName(ev_tab)," does not correspond to a recognized mode ",&
        "(IMAGE.EL, TIME.EL, RIMAGE.EL, CTIME.EL, BURST.EL, SPECT.EL, HTR.EL)."
       call error("getMode01")
    end select
    write(errstr,*) "DATATYPE = ", trim(keyword)
    call message(AppMsg,Verbose)

  end subroutine getMode

! Set to 1 a keyword stating that an action governed by a boolean parameter
! was performed

  subroutine keywordDone(in_tab, taskname, paramname)

! in_tab:    handle to the table where the keyword will be written
! taskname:  name of the calling task (written in the comment)
! paramname: name of the boolean parameter

    type(BlockT), intent(in) :: in_tab
    character(len=*), intent(in) :: taskname, paramname

    character(len=8) :: keyword
    character(len=1024) :: upparam
    integer :: nchar

    nchar = min(8,len(paramname))
    keyword = "        "
    upparam = changeCase(paramname,2)
    keyword(1:nchar) = upparam(1:nchar)

    upparam = taskname // " was applied with " // trim(paramname) // "=yes"
    call setAttribute(in_tab, keyword, 1, comment=upparam)

  end subroutine keywordDone

! Remove a keyword written with keywordDone

  subroutine keywordRemove(in_tab, paramname)

! in_tab:    handle to the table where the keyword will be set to 0
! paramname: name of the boolean parameter

    type(BlockT), intent(in) :: in_tab
    character(len=*), intent(in) :: paramname

    character(len=8) :: keyword
    character(len=1024) :: upparam
    integer :: nchar

    nchar = min(8,len(paramname))
    keyword = "        "
    upparam = changeCase(paramname,2)
    keyword(1:nchar) = upparam(1:nchar)

    call setAttribute(in_tab, keyword, 0)

  end subroutine keywordRemove

! Test using keywords whether an action governed by a boolean parameter
! was already performed

  function wasDone(in_tab,paramname) result(done)

! in_tab:    handle to the table where to look for the keyword
! paramname: name of the boolean parameter

    type(BlockT), intent(in) :: in_tab
    character(len=*), intent(in) :: paramname
    logical :: done

    character(len=8) :: keyword
    character(len=1024) :: upparam
    integer :: nchar
    integer(kind=Int32) :: val

    nchar = min(8,len(paramname))
    keyword = "        "
    upparam = changeCase(paramname,2)
    keyword(1:nchar) = upparam(1:nchar)
    done = hasAttribute(in_tab,keyword)
    if (done) then  
       val = int32Attribute(in_tab,keyword)
       done = val > 0
    endif

  end function wasDone

! Check that two files share a number of attributes

  function equalKeywords(handle1,handle2,keywList,strict,onwarn) result(compat)

! handle1    : handle to the first table or set
! handle2    : handle to the second table or data set
! keywList   : array of keywords to test for compatibility
! strict     : set to true if need to check existence also (default is true)
! onwarn     : set to true if warnings are to be sent (default is true)
!            : if false then messages are sent instead

    type(AttributableT), intent(in) :: handle1, handle2
    character(len=*), dimension(:), intent(in) :: keywList
    logical, intent(in), optional :: strict, onwarn
    logical :: compat

! Tolerance for 'real' equality

    real (kind=single), parameter :: epsilon  = 1.e-5

    character(len=80) :: inValue, outValue
    integer :: nk, nchars, inInt, outInt
    real (kind=single) :: inReal, outReal, denom
    logical :: strict2, onwarn2

    if (present(strict)) then
       strict2 = strict
    else
       strict2 = .true.
    endif
    if (present(onwarn)) then
       onwarn2 = onwarn
    else
       onwarn2 = .true.
    endif

    compat = .true.
    do nk = 1,size(keywList)
       if (strict2) then
          compat = compat .and. hasAttribute(handle1,keywList(nk))
          if (.not. compat) then
             write(errstr,*) "Keyword ", trim(keywList(nk)),  &
                " is not present in ", trim(qualifiedName(handle1))
             if (onwarn2) then
                call warning("equalKeywords11")
             else
                call message(AppMsg,Sparse)
             endif
             return
          endif
          compat = compat .and. hasAttribute(handle2,keywList(nk))
          if (.not. compat) then
             write(errstr,*) "Keyword ", trim(keywList(nk)),  &
                " is not present in ", trim(qualifiedName(handle2))
             if (onwarn2) then
                call warning("equalKeywords12")
             else
                call message(AppMsg,Sparse)
             endif
             return
          endif
       else if (.not. (hasAttribute(handle1,keywList(nk)) .and.  &
                       hasAttribute(handle2,keywList(nk)))) then
          cycle
       endif

       select case(datatype(attribute(handle1,keywList(nk))))
       case (Boolean_attribute, String_attribute)
           inValue = stringAttribute(handle1,keywList(nk))
          outValue = stringAttribute(handle2,keywList(nk))
          nchars = min(len_trim(inValue),len_trim(outValue))
          compat = compat .and. (inValue(1:nchars) == outValue(1:nchars))
       case (Integer_attribute)
          select case(datatype(attribute(handle2,keywList(nk))))
          case (Integer_attribute)
              inInt = int32Attribute(handle1,keywList(nk))
             outInt = int32Attribute(handle2,keywList(nk))
             compat = compat .and. (inInt == outInt)
          case default
             compat = .false.
             write(errstr,*) "Incompatible types for attribute ",  &
                             trim(keywList(nk)), " in ",  &
                             trim(qualifiedName(handle1)), " and ",  &
                             trim(qualifiedName(handle2))
             call warning("equalKeywords15")
          end select
       case (Real_attribute)
          select case(datatype(attribute(handle2,keywList(nk))))
          case (Real_attribute)
              inReal = real32Attribute(handle1,keywList(nk))
             outReal = real32Attribute(handle2,keywList(nk))
             denom = max(abs(inReal+outReal),tiny(epsilon))
             compat = compat .and. abs(inReal-outReal)/denom < epsilon
          case default
             compat = .false.
             write(errstr,*) "Incompatible types for attribute ",  &
                             trim(keywList(nk)), " in ",  &
                             trim(qualifiedName(handle1)), " and ",  &
                             trim(qualifiedName(handle2))
             call warning("equalKeywords15")
          end select
       case default
          write(errstr,*) "Unknown type ",  &
             datatype(attribute(handle1,keywList(nk))), &
             " for attribute ", trim(keywList(nk)), " in ",  &
             trim(qualifiedName(handle1)), ". Comparison unreliable."
          call warning("equalKeywords14")
       end select

        inValue = stringAttribute(handle1,keywList(nk))
       outValue = stringAttribute(handle2,keywList(nk))
       if (.not. compat) then
          write(errstr,*) "Keyword ", trim(keywList(nk)),  &
             " not identical in both files ", trim(inValue),  &
             " in ", trim(qualifiedName(handle1)), " vs ", trim(outValue),  &
             " in ", trim(qualifiedName(handle2))
          if (onwarn2) then
             call warning("equalKeywords13")
          else
             call message(AppMsg,Sparse)
          endif
          return
! Trick to write the keywords once in evlistcomb
       else if (qualifiedName(handle1) == qualifiedName(handle2)) then
          write(errstr,*) trim(keywList(nk)), " = ", trim(inValue)
          call message(AppMsg,Sparse)
       endif
    enddo

  end function equalKeywords

! Copy general keywords from extension to primary header

  subroutine putPrimaryKeywords(fr_tab)

! fr_tab    : Handle to the input block

    type(BlockT), intent(in)   :: fr_tab

! Names of the keywords to be copied into the primary header from the extension

    integer, parameter :: num_keyw_prim = 6
    character(len=8), dimension(num_keyw_prim), parameter :: &
       name_keyw_prim = (/"TELESCOP","INSTRUME", "OBS_ID  ", "EXP_ID  ",  &
                          "DATE-OBS","DATE-END"/)
  
    character(len=8) :: keyword
    integer :: ncur

    call message(AppMsg,Verbose,"Entering putPrimaryKeywords")

    do ncur = 1, num_keyw_prim
       keyword = name_keyw_prim(ncur)
       if (hasAttribute(fr_tab,keyword)) then
          call setAttribute(parent(fr_tab), attribute(fr_tab,keyword))
       else
          write(errstr,*) "keyword ", keyword, " not found in ",  &
                          trim(qualifiedName(fr_tab))
          call warning("putPrimaryKeywords10")
       endif
    enddo
   end subroutine putPrimaryKeywords

! Read the size of a parameter list

  function sizeListParam( paramlist ) result( numlist )

! paramlist  : name of the string list parameter
! numlist    : number of strings in list

    character(len=*), intent(in)               :: paramlist
    integer                                    :: numlist

    numlist = parameterCount(paramlist)
    if (numlist < 0) then
      write(errstr,*) "Incorrect parameter list :", trim(paramlist)
      call warning("sizeListParam10")
      numlist = 0  
    endif

  end function sizeListParam

! Read a parameter list (strings only)

  subroutine readListParam(paramlist,list,numlist)

! paramlist  : name of the string list parameter
! list       : list of strings
! numlist    : number of strings in list

    character(len=*), intent(in)                     :: paramlist
    character(len=*), dimension(:), intent(inout)    :: list
    integer, intent(in)                              :: numlist
   
    integer             :: nn
    character(len=1024) :: param

    write(errstr,*) numlist, " input names for parameter :",paramlist
    call message(AppMsg,Sparse)
    do nn=1,numlist
       param = stringParameter(paramlist,nn-1)
       if ( len_trim(param) > len(list(nn)) ) then
          write(errstr,*) "Parameter ",trim(param), " (longer than ", & 
             len(list(nn)), " char.) is truncated"
          call warning("readListParam10")
       endif
       list(nn) = param
       write(errstr,*) nn," : ",list(nn)
       call message(AppMsg,Noisy)
    enddo

  end subroutine readListParam

end module emutils_mod
